from .mri import (  # noqa: F401
    Kspace,
    Rda,
)
from .pet import (
    PetRawData,
    PetListMode,
    PetSinogram,
    PetCountRate,
    PetNormalisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData,
    Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogram,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogramSeries,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetParameterisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData,
)

__all__ = [
    "Kspace",
    "Rda",
    "PetRawData",
    "PetListMode",
    "PetSinogram",
    "PetCountRate",
    "PetNormalisation",
    "Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData",
    "Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData",
    "Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode",
    "Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram",
    "Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogram",
    "Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogramSeries",
    "Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate",
    "Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation",
    "Vnd_Siemens_Biograph128Vision_Vr20b_PetParameterisation",
    "Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData",
]



import io
import typing as ty
from collections import defaultdict
import sys
from pathlib import Path
from fileformats.generic import TypedSet
from fileformats.core import mtime_cached_property, validated_property, extra
from fileformats.core.exceptions import FormatMismatchError
from fileformats.core.mixin import WithMagicNumber
from fileformats.medimage.dicom import get_dicom_tag
from fileformats.core.io import BinaryIOWindow
from .base import (
    PetRawData,
    PetPhysio,
    PetListMode,
    PetSinogram,
    PetCountRate,
    PetNormalisation,
    PetParameterisation,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if ty.TYPE_CHECKING:
    import pydicom


class Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData(PetRawData):
    """PET raw data format as produced by Siemens Biograph 128 Vision. It is used to
    store a range of raw PET data such as list-mode, calibration and sinogram files.

    Consists of a block of raw data, followed by a DICOM header, an int4 containing
    the size of the DICOM header, and then the number b'LARGE_PET_LM_RAWDATA'.
    """

    ext = ".ptd"
    expected_image_type: str = ""  # For sub-classes to specify the expected image tag
    IMAGE_TYPE_TAG = (0x0008, 0x0008)

    def read_tag(self, tag: ty.Tuple[int, int]) -> ty.Union[str, bytes, None]:
        with self.open() as f:
            window = BinaryIOWindow(
                f,  # type: ignore[arg-type]
                *self.dicom_header_limits,
            )
            value = get_dicom_tag(window, tag)
        return value

    @validated_property
    def image_type(self) -> str:
        image_type_seq: ty.List[str] = self._image_type_seq
        if self.expected_image_type:
            if (
                not isinstance(image_type_seq, list)
                or len(image_type_seq) != 3
                or image_type_seq[:2] != ["ORIGINAL", "PRIMARY"]
            ):
                raise FormatMismatchError(
                    f"Image type of {self} ({image_type_seq!r}) does not match expected "
                    f"form ['ORIGINAL', 'PRIMARY', *]"
                )
            image_type = image_type_seq[2]
            if self.expected_image_type != image_type:
                raise FormatMismatchError(
                    f"Image type of {self} ({image_type!r}) does not match expected "
                    f"({self.expected_image_type!r})"
                )
            return image_type
        assert isinstance(self._image_type_seq[-1], str)
        return self._image_type_seq[-1]

    @mtime_cached_property
    def _image_type_seq(self) -> ty.List[str]:
        value = self.read_tag(self.IMAGE_TYPE_TAG)
        assert isinstance(value, str)
        return value.split("\\")

    @mtime_cached_property
    def dicom_header_limits(self) -> ty.Tuple[int, int]:
        """Try either method to determine the dicom header limits. If both fail, raise an error"""
        try:
            return Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData(
                self.fspaths
            ).dicom_header_limits
        except FormatMismatchError:
            try:
                return Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData(
                    self.fspaths
                ).dicom_header_limits
            except FormatMismatchError:
                raise FormatMismatchError(
                    f"File {self.fspath!r} does not contain a valid DICOM header"
                ) from None

    @extra
    def load_pydicom(self, **kwargs: ty.Any) -> "pydicom.Dataset":
        """Reads any metadata associated with the fileset and returns it as a dict

        Parameters
        ----------
        **kwargs : Any
            any format-specific keyword arguments to pass to the metadata reader

        Returns
        -------
        pydicom.Dataset
            the DICOM dataset containing the metadata loaded by PyDICOM
        """
        raise NotImplementedError


class Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData(
    WithMagicNumber, Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData
):
    """PET raw data format as produced by Siemens Biograph 128 Vision. It is used to
    store a range of raw PET data such as list-mode, calibration and sinogram files.

    Consists of a block of raw data, followed by a DICOM header, an int4 containing
    the size of the DICOM header, and then the number b'LARGE_PET_LM_RAWDATA'.
    """

    ext = ".ptd"
    magic_number = b"LARGE_PET_LM_RAWDATA"
    magic_number_offset = -len(magic_number)  # magic number is at end of file

    # Size in bytes of the integer that defines the size of the dicom header
    sizeof_dcm_hdr_size_int: int = 4
    # Offset from the end of the file where the magic number and header size integer
    dcm_hdr_size_int_offset: int = magic_number_offset - sizeof_dcm_hdr_size_int

    @mtime_cached_property
    def dicom_header_size(self) -> int:
        with self.open() as f:
            f.seek(self.dcm_hdr_size_int_offset, io.SEEK_END)
            dcm_hdr_size_bytes: bytes = f.read(self.sizeof_dcm_hdr_size_int)
            return int.from_bytes(dcm_hdr_size_bytes, "little")

    @validated_property
    def dicom_header_offset(self) -> int:
        dcm_hdr_size: int = self.dicom_header_size
        return -dcm_hdr_size + self.dcm_hdr_size_int_offset

    @property
    def dicom_header_limits(self) -> ty.Tuple[int, int]:  # type: ignore[override]
        """Returns the start and end of the DICOM header in the file"""
        return (
            self.dicom_header_offset,
            self.dicom_header_offset + self.dicom_header_size,
        )


class Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode(
    Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData, PetListMode
):
    expected_image_type = "PET_LISTMODE"


class Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram(
    Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData, PetSinogram
):
    "histogrammed projection data in a reconstruction-friendly format"

    expected_image_type = "PET_EM_SINOGRAM"


class Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogram(
    Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData, PetSinogram
):
    "histogrammed projection data in a reconstruction-friendly format"

    expected_image_type = "PET_SINO_DYNAMIC"


class Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate(
    Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData, PetCountRate
):
    "number of prompt/random/single events per unit time"

    expected_image_type = "PET_COUNTRATE"


class Vnd_Siemens_Biograph128Vision_Vr20b_PetParameterisation(
    Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData, PetParameterisation
):
    "number of prompt/random/single events per unit time"

    expected_image_type = "PET_REPLAY_PARAM"


class Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation(
    Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData, PetNormalisation
):
    "normalisation scan or the current cross calibration factor"

    expected_image_type = "PET_CALIBRATION"


class Vnd_Siemens_Biograph128Vision_Vr20b_PetPhysio(
    Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData, PetPhysio
):
    "normalisation scan or the current cross calibration factor"

    expected_image_type = "PET_PHYSIO"


class Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogramSeries(TypedSet):
    "Series of sinogram images"

    content_types = (Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogram,)

    @classmethod
    def from_paths(
        cls,
        fspaths: ty.Iterable[Path],
        common_ok: bool = False,
        **kwargs: ty.Any,
    ) -> ty.Tuple[ty.Set[Self], ty.Set[Path]]:
        """Separates a list of DICOM files into separate series from the file-system
        paths

        Parameters
        ----------
        fspaths : ty.Iterable[Path]
            the fspaths pointing to the DICOM files
        common_ok : bool, optional
            included to match the signature of the overridden method, but ignored as each
            dicom should belong to only one series.
        specific_tags : ty.Optional[TagListType], optional
            the DICOM tags to read from the files. If None, the default tags will be
            read
        **kwargs : ty.Any
            additional keyword arguments to passed through to the DicomImage constructor

        Returns
        -------
        tuple[set[DicomSeries], set[Path]]
            the found dicom series objects and any unrecognised file paths
        """
        (
            sinograms,
            remaining,
        ) = Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogram.from_paths(
            fspaths, common_ok=common_ok, **kwargs
        )
        series_dict = defaultdict(list)
        for sinogram in sinograms:
            series_dict[tuple(sinogram.read_tag(t) for t in cls.ID_TAGS)].append(
                sinogram
            )
        return set([cls(d.fspath for d in s) for s in series_dict.values()]), remaining

    @mtime_cached_property
    def contents(self) -> ty.List[Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram]:
        return sorted(TypedSet.contents.__get__(self), key=pet_rd_sort_key)

    ID_TAGS = (
        (0x0020, 0x000D),
        (0x0020, 0x0011),
    )  # "StudyInstanceUID", "SeriesNumber"


def pet_rd_sort_key(ptd: Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData) -> str:
    """Sorts DICOM objects by SOPInstanceUID"""
    acquisition_time = ptd.read_tag((0x0008, 0x0032))
    assert isinstance(acquisition_time, str)
    return acquisition_time


class Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData(
    WithMagicNumber, Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData
):
    """PET CT raw data format as produced by Siemens Biograph 128 Vision. It is used to
    store a range PETCT_SPL type files

    Consists of a block of dummy, followed by a DICOM header.
    """

    ext = ".ptd"
    magic_number = b"dummy data"
    expected_image_type: str = "PETCT_SPL"

    magic_dicom_end = b"END!"

    @validated_property
    def dicom_header_limits(self) -> ty.Tuple[int, int]:  # type: ignore[override]
        """Returns the start and end indices of the DICOM data embedded within the
        file. The DICOM data starts immediately after the magic number and is delimited
        by a magic sequence 'END!'"""
        start = len(self.magic_number)
        end = None
        with self.open() as f:
            while True:
                chunk = f.read(1024)
                if not chunk:
                    break
                offset = chunk.find(self.magic_dicom_end)
                if offset != -1:
                    end = f.tell() - len(chunk) + offset
                    return (start, end)
        raise FormatMismatchError(
            f"Magic sequence {self.magic_dicom_end!r} delimiting DICOM data "
            f"not found in file PETCT SPL({str(self)!r})"
        )



import typing as ty
from pathlib import Path
import pydicom
import sys
from collections import Counter
from fileformats.core import SampleFileGenerator
from medimages4tests.dummy.raw.pet.siemens.biograph_vision.vr20b.pet_listmode import (
    get_data as get_pet_listmode_data,
)
from medimages4tests.dummy.raw.pet.siemens.biograph_vision.vr20b.pet_countrate import (
    get_data as get_pet_countrate_data,
)
from medimages4tests.dummy.raw.pet.siemens.biograph_vision.vr20b.pet_em_sino import (
    get_data as get_pet_sinogram_data,
)
from medimages4tests.dummy.raw.pet.siemens.biograph_vision.vr20b.pet_calibration import (
    get_data as get_pet_calibration_data,
)
from medimages4tests.dummy.raw.pet.siemens.biograph_vision.vr20b.pet_dynamics_sino import (
    get_data as get_pet_dynamics_sino_data,
)
from medimages4tests.dummy.raw.pet.siemens.biograph_vision.vr20b.pet_replay_param import (
    get_data as get_pet_replay_param_data,
)
from medimages4tests.dummy.raw.pet.siemens.biograph_vision.vr20b.petct_spl import (
    get_data as get_petct_spl_data,
)
from fileformats.core import extra_implementation, FileSet
from fileformats.medimage.dicom import DicomImage
from fileformats.medimage.raw import (
    Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData,
    Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogramSeries,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetParameterisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData,
)
from fileformats.core.io import BinaryIOWindow

if sys.version_info >= (3, 9):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

TagListType: TypeAlias = ty.Union[
    ty.List[int],
    ty.List[str],
    ty.List[ty.Tuple[int, int]],
    ty.List[pydicom.tag.BaseTag],
]


@extra_implementation(FileSet.read_metadata)
def siemens_pet_raw_data_read_metadata(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData,
    specific_tags: ty.Optional[TagListType] = None,
    **kwargs: ty.Any,
) -> ty.Mapping[str, ty.Any]:

    with pet_raw_data.open() as f:
        window = BinaryIOWindow(
            f,  # type: ignore[arg-type]
            pet_raw_data.dicom_header_offset,
            pet_raw_data.dcm_hdr_size_int_offset,
        )
        dcm = pydicom.dcmread(window, specific_tags=specific_tags)
    return DicomImage.pydicom_to_dict(dcm)


@extra_implementation(Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData.load_pydicom)
def siemens_pet_raw_data_load_pydicom(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_LargePetRawData,
    specific_tags: ty.Optional[TagListType] = None,
    **kwargs: ty.Any,
) -> pydicom.Dataset:

    with pet_raw_data.open() as f:
        window = BinaryIOWindow(
            f,  # type: ignore[arg-type]
            pet_raw_data.dicom_header_offset,
            pet_raw_data.dcm_hdr_size_int_offset,
        )
        dcm = pydicom.dcmread(window, specific_tags=specific_tags)
    return dcm


@extra_implementation(FileSet.read_metadata)
def siemens_petct_raw_data_read_metadata(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData,
    specific_tags: ty.Optional[TagListType] = None,
    **kwargs: ty.Any,
) -> ty.Mapping[str, ty.Any]:

    with pet_raw_data.open() as f:
        window = BinaryIOWindow(
            f,  # type: ignore[arg-type]
            *pet_raw_data.dicom_header_limits,
        )
        dcm = pydicom.dcmread(window, specific_tags=specific_tags)
    return DicomImage.pydicom_to_dict(dcm)


@extra_implementation(Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData.load_pydicom)
def siemens_petct_raw_data_load_pydicom(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData,
    specific_tags: ty.Optional[TagListType] = None,
    **kwargs: ty.Any,
) -> pydicom.Dataset:

    with pet_raw_data.open() as f:
        window = BinaryIOWindow(
            f,  # type: ignore[arg-type]
            *pet_raw_data.dicom_header_limits,
        )
        dcm = pydicom.dcmread(window, specific_tags=specific_tags)
    return dcm


@extra_implementation(FileSet.read_metadata)
def siemens_pet_dynamic_sinogram_series_read_metadata(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogramSeries,
    specific_tags: ty.Optional[TagListType] = None,
    **kwargs: ty.Any,
) -> ty.Mapping[str, ty.Any]:

    # Collated DICOM headers across series
    collated: ty.Dict[str, ty.Any] = {}
    key_repeats: ty.Counter[str] = Counter()
    varying_keys = set()
    # We use the "contents" property implementation in TypeSet instead of the overload
    # in DicomCollection because we don't want the metadata to be read ahead of the
    # the `select_metadata` call below

    for ptd in pet_raw_data.contents:
        for key, val in ptd.metadata.items():
            try:
                prev_val = collated[key]
            except KeyError:
                collated[
                    key
                ] = val  # Insert initial value (should only happen on first iter)
                key_repeats.update([key])
            else:
                if key in varying_keys:
                    collated[key].append(val)
                # Check whether the value is the same as the values in the previous
                # images in the series
                elif val != prev_val:
                    collated[key] = [prev_val] * key_repeats[key] + [val]
                    varying_keys.add(key)
                else:
                    key_repeats.update([key])
    return collated


@extra_implementation(FileSet.generate_sample_data)
def siemens_pet_listmode_generate_sample_data(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return get_pet_listmode_data(out_dir=generator.dest_dir)  # type: ignore[no-any-return]


@extra_implementation(FileSet.generate_sample_data)
def siemens_pet_countrate_generate_sample_data(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return get_pet_countrate_data(out_dir=generator.dest_dir)  # type: ignore[no-any-return]


@extra_implementation(FileSet.generate_sample_data)
def siemens_pet_sinogram_generate_sample_data(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return get_pet_sinogram_data(out_dir=generator.dest_dir)  # type: ignore[no-any-return]


@extra_implementation(FileSet.generate_sample_data)
def siemens_pet_dynamics_sino_generate_sample_data(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogramSeries,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return get_pet_dynamics_sino_data(out_dir=generator.dest_dir)  # type: ignore[no-any-return]


@extra_implementation(FileSet.generate_sample_data)
def siemens_pet_normalisation_generate_sample_data(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return get_pet_calibration_data(out_dir=generator.dest_dir)  # type: ignore[no-any-return]


@extra_implementation(FileSet.generate_sample_data)
def siemens_petct_spl_generate_sample_data(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return get_petct_spl_data(out_dir=generator.dest_dir)  # type: ignore[no-any-return]


@extra_implementation(FileSet.generate_sample_data)
def siemens_pet_parameterisation_generate_sample_data(
    pet_raw_data: Vnd_Siemens_Biograph128Vision_Vr20b_PetParameterisation,
    generator: SampleFileGenerator,
) -> ty.List[Path]:
    return get_pet_replay_param_data(out_dir=generator.dest_dir)  # type: ignore[no-any-return]



from pathlib import Path
from fileformats.core import from_paths

# from fileformats.medimage.raw import Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData
from fileformats.medimage.raw import (
    Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogramSeries,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetParameterisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData,
)

Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData(
    "/Users/tclose/Data/tbp/PET_Raw_Data_0602/PHYSICS_SUV_TEST_PVE_F18.PT.PET_Onco_(Adult).602.PETCT_SPL.2023.07.27.08.49.11.676000.2.0.2761828.ptd"
)

filesets = from_paths(
    Path("/Users/tclose/Data/tbp/").glob("**/*.ptd"),
    Vnd_Siemens_Biograph128Vision_Vr20b_PetDynamicSinogramSeries,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetParameterisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCtSplRawData,
)


# ptd = Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData(
#     Path(
#         "/Users/tclose/Data/tbp/PET_Raw_Data_0602/"
#         "PHYSICS_SUV_TEST_PVE_F18.PT.PET_Onco_(Adult).602.PET_LISTMODE.2023.07.27.08.49.11.676000.2.0.2761856.ptd"
#     ).expanduser()
# )

# mdata = ptd.metadata

assert filesets



import numpy as np
import typing  # noqa: F401
import numpy.typing
from fileformats.core import extra_implementation
from fileformats.medimage import DwiEncoding, Bval, Bvec
from fileformats.medimage.diffusion import EncodingArrayType


@extra_implementation(Bval.read_array)
def bval_read_array(bval: Bval) -> EncodingArrayType:
    return np.asarray([float(ln) for ln in bval.read_contents().split()])


@extra_implementation(DwiEncoding.read_array)
def bvec_read_array(bvec: Bvec) -> EncodingArrayType:
    bvals = bvec.b_values_file.read_array()
    directions = np.asarray(
        [[float(x) for x in ln.split()] for ln in bvec.read_contents().splitlines()]
    ).T
    return np.concatenate((directions, bvals.reshape((-1, 1))), axis=1)
